"""Utilities for generating SQL while avoiding SQL injection vulns"""

from firepit.validate import validate_name
from firepit.validate import validate_path

import re

COMP_OPS = ['=', '<>', '!=', '<', '>', '<=', '>=', 'LIKE', 'IN', 'IS', 'IS NOT']
PRED_OPS = ['AND', 'OR']
JOIN_TYPES = ['INNER', 'OUTER', 'LEFT OUTER', 'CROSS']
AGG_FUNCS = ['COUNT', 'SUM', 'MIN', 'MAX', 'AVG', 'NUNIQUE']


class InvalidComparisonOperator(Exception):
    pass


class InvalidPredicateOperator(Exception):
    pass


class InvalidJoinOperator(Exception):
    pass


class InvalidAggregateFunction(Exception):
    pass


class InvalidQuery(Exception):
    pass


class Predicate:
    """Simple row value predicate"""

    def __init__(self, lhs, op, rhs):
        if op not in COMP_OPS:
            raise InvalidComparisonOperator(op)
        if rhs is None:
            rhs = 'NULL'
        if lhs.endswith('[*]'):  # STIX list property
            lhs = lhs[:-3]
            if rhs.lower() != 'null':
                rhs = f"%{rhs}%"  # wrap with SQL wildcards since list is encoded as string
                if op == '=':
                    op = 'LIKE'
                elif op == '!=':
                    op = 'NOT LIKE'
        validate_path(lhs)
        self.lhs = lhs
        self.op = op
        self.rhs = rhs
        if self.rhs in ['null', 'NULL']:
            self.values = ()
            if op not in ['=', '!=', '<>', 'IS', 'IS NOT']:
                raise InvalidComparisonOperator(op)  # Maybe need different exception here?
        elif isinstance(self.rhs, (list, tuple)):
            self.values = tuple(self.rhs)
        else:
            self.values = (self.rhs, )

    def render(self, placeholder):
        if self.rhs in ['null', 'NULL']:
            if self.op in ['!=', '<>']:
                text = f'("{self.lhs}" IS NOT NULL)'
            elif self.op == '=':
                text = f'("{self.lhs}" IS NULL)'
            else:
                raise InvalidComparisonOperator(self.op)
        elif self.op == 'IN':
            phs = ', '.join([placeholder] * len(self.rhs))
            text = f'("{self.lhs}" {self.op} ({phs}))'
        else:
            text = f'("{self.lhs}" {self.op} {placeholder})'
        return text


class Filter:
    """Alternative SQL WHERE clause"""

    OR = ' OR '
    AND = ' AND '

    def __init__(self, preds, op=AND):
        self.preds = preds
        self.op = op
        self.values = ()
        for pred in self.preds:
            self.values += pred.values

    def render(self, placeholder):
        pred_list = []
        for pred in self.preds:
            pred_list.append(pred.render(placeholder))
        result = self.op.join(pred_list)
        if self.op == Filter.OR:
            return f'({result})'
        return result


class Order:
    """SQL ORDER BY clause"""

    ASC = 'ASC'
    DESC = 'DESC'

    def __init__(self, cols):
        self.cols = []
        for col in cols:
            if isinstance(col, tuple):
                validate_path(col[0])
                self.cols.append(col)
            elif isinstance(col, str):
                validate_path(col)
                self.cols.append((col, Order.ASC))

    def render(self, placeholder):
        col_list = []
        for col in self.cols:
            col_list.append(f'"{col[0]}" {col[1]}')
        return ', '.join(col_list)


class Projection:
    """SQL SELECT (really projection - pick column subset) clause"""

    def __init__(self, cols):
        for col in cols:
            validate_path(col)
        self.cols = cols

    def render(self, placeholder):
        return ', '.join([f'"{col}"' for col in self.cols])


class Table:
    """SQL Table selection"""

    def __init__(self, name):
        validate_name(name)
        self.name = name

    def render(self, placeholder):
        return self.name


class Group:
    """SQL GROUP clause"""

    def __init__(self, cols):
        for col in cols:
            validate_path(col)
        self.cols = cols

    def render(self, placeholder):
        return ', '.join([f'"{col}"' for col in self.cols])


class Aggregation:
    """Aggregate rows"""

    def __init__(self, aggs):
        self.aggs = []
        for agg in aggs:
            if isinstance(agg, tuple):
                if len(agg) == 3:
                    func, col, alias = agg
                elif len(agg) == 2:
                    func, col = agg
                    alias = None
                if func.upper() not in AGG_FUNCS:
                    raise InvalidAggregateFunction(func)
                if col is not None and col != '*':
                    validate_path(col)
                self.aggs.append((func, col, alias))
            else:
                raise TypeError('expected aggregation tuple but received ' + str(type(agg)))
        self.group_cols = []  # Filled in by Query

    def render(self, placeholder):
        exprs = [f'"{col}"' for col in self.group_cols]
        for agg in self.aggs:
            mod = ''
            func, col, alias = agg
            if func.upper() == 'NUNIQUE':
                func = 'COUNT'
                mod = 'DISTINCT '
            if not col:
                col = '*'
            if col == '*':
                expr = f'{func}({mod}{col})'  # No quotes for *
            else:
                expr = f'{func}({mod}"{col}")'
            if not alias:
                alias = func.lower()
            expr += f' AS "{alias}"'
            exprs.append(expr)
        return ', '.join(exprs)


class Offset:
    """SQL row offset"""

    def __init__(self, num):
        self.text = f'{num}'

    def render(self, placeholder):
        return self.text


class Limit:
    """SQL row count"""

    def __init__(self, num):
        self.text = f'{num}'

    def render(self, placeholder):
        return self.text


class Count:
    """Count the rows in a result set"""

    def __init__(self):
        pass

    def render(self, placeholder):
        return 'COUNT(*) AS "count"'


class Unique:
    """Reduce the rows in a result set to unique tuples"""

    def __init__(self):
        pass

    def render(self, placeholder):
        return 'SELECT DISTINCT *'


class CountUnique:
    """Unique count of the rows in a result set"""

    def __init__(self, cols=None):
        for col in cols or []:
            validate_path(col)
        self.cols = cols

    def render(self, placeholder):
        if self.cols:
            cols = ', '.join([f'"{col}"' for col in self.cols])
            return f'COUNT(DISTINCT {cols}) AS "count"'
        return 'COUNT(*) AS "count"'


class Join:
    """Join 2 tables"""

    def __init__(self, name, left_col, op, right_col, how='INNER'):
        validate_name(name)
        validate_path(left_col)
        validate_path(right_col)
        if how.upper() not in JOIN_TYPES:
            raise InvalidJoinOperator(how)
        self.prev_name = None
        self.name = name
        self.left_col = left_col
        self.op = op
        self.right_col = right_col
        self.how = how

    def render(self, placeholder):
        # Assume there's a FROM before this?
        return (f'{self.how.upper()} JOIN "{self.name}"'
                f' ON "{self.prev_name}"."{self.left_col}"'
                f' {self.op} "{self.name}"."{self.right_col}"')


class Query:
    def __init__(self, arg=None):
        self.stages = []
        if isinstance(arg, str):
            self.append(Table(arg))
        elif isinstance(arg, list):
            for stage in arg:
                self.append(stage)

    def last_stage(self):
        return self.stages[-1] if self.stages else None

    def append(self, stage):
        if isinstance(stage, Aggregation):
            # If there's already a Projection, that's an error
            for prev in self.stages:
                if isinstance(prev, Projection):
                    raise InvalidQuery('cannot have Aggregation after Projection')
            if self.stages:
                last = self.last_stage()
                if isinstance(last, Group):
                    stage.group_cols = last.cols  # Copy grouped columns
        elif isinstance(stage, Join):
            # Need to look back and grab previous table name
            last = self.last_stage()
            if isinstance(last, (Table, Join)):
                stage.prev_name = last.name
            else:
                raise InvalidQuery('Join must follow Table or Join')
        elif isinstance(stage, Count):
            # See if we can combine with previous stages
            last = self.last_stage()
            if isinstance(last, Unique):
                self.stages.pop(-1)
                cols = None
                if self.stages:
                    last = self.last_stage()
                    if isinstance(last, Projection):
                        proj = self.stages.pop(-1)
                        cols = proj.cols
                stage = CountUnique(cols)
        elif isinstance(stage, CountUnique):
            # See if we can combine with previous stages
            last = self.last_stage()
            if isinstance(last, Projection):
                self.stages.pop(-1)
                stage.cols = last.cols
        self.stages.append(stage)

    def render(self, placeholder):
        # TODO: detect missing table
        query = ''
        values = ()
        prev = None  # TODO: Probably need state machine here
        for stage in self.stages:
            text = stage.render(placeholder)
            if isinstance(stage, Table):
                query = f'FROM "{text}"'
            elif isinstance(stage, Projection):
                query = f'SELECT {text} {query}'
            elif isinstance(stage, Filter):
                values += stage.values
                if isinstance(prev, Aggregation):
                    keyword = 'HAVING'
                elif isinstance(prev, Filter):
                    keyword = 'AND'
                else:
                    keyword = 'WHERE'
                query = f'{query} {keyword} {text}'
            elif isinstance(stage, Group):
                query = f'{query} GROUP BY {text}'
            elif isinstance(stage, Aggregation):
                query = f'SELECT {text} {query}'
            elif isinstance(stage, Order):
                query = f'{query} ORDER BY {text}'
            elif isinstance(stage, Limit):
                query = f'{query} LIMIT {text}'
            elif isinstance(stage, Offset):
                query = f'{query} OFFSET {text}'
            elif isinstance(stage, Count):
                if isinstance(prev, Unique):
                    # Should have already been combined, so we should never hit this
                    query = f'SELECT {text} FROM ({query}) AS tmp'
                else:
                    query = f'SELECT {text} {query}'
            elif isinstance(stage, Unique):
                if isinstance(prev, Projection):
                    query = re.sub(r'^SELECT ', 'SELECT DISTINCT ', query)
                else:
                    query = f'SELECT DISTINCT * {query}'
            elif isinstance(stage, Join):
                query = f'{query} {text}'
            elif isinstance(stage, CountUnique):
                if stage.cols:
                    query = f'SELECT {text} {query}'
                else:
                    query = f'SELECT {text} FROM (SELECT DISTINCT * {query}) AS tmp'
            prev = stage
        # If there's no projection...
        if not query.startswith('SELECT'):  # Hacky
            query = 'SELECT * ' + query
        return query, values



"""Utility functions for STIX properties"""

import re


# A regex to grab the last piece of a STIX path
last_re = re.compile(r'.*[\.:]([a-z]*)')


def get_last(prop):
    return last_re.sub(r'\g<1>', prop)


def primary_prop(sco_type):
    """Returns the "primary" property name for each SCO type"""
    prop = 'value'  # Generic catchall
    if sco_type == 'user-account':
        prop = 'user_id'
    elif sco_type in ['file', 'mutex', 'process',
                      'software', 'windows-registry-value-type']:
        prop = 'name'
    elif sco_type == 'directory':
        prop = 'path'
    elif sco_type == 'autonomous-system':
        prop = 'number'
    elif sco_type == 'windows-registry-key':
        prop = 'key'
    elif sco_type == 'x509-certificate':
        prop = 'serial_number'
    return prop


def auto_agg(sco_type, prop, col_type):
    """Infer an aggregation function based on column name and type"""

    agg = auto_agg_tuple(sco_type, prop, col_type)
    if not agg:
        return None

    func, col, alias = agg

    if len(alias) > 63:
        # PostgreSQL has a limit of 63 chars per identifier
        return None

    # Special case for NUNIQUE (which is not SQL)
    if func == 'NUNIQUE':
        return f'COUNT(DISTINCT "{col}") AS "{alias}"'
    return f'{func}("{col}") AS "{alias}"'


def auto_agg_tuple(sco_type, prop, col_type):
    """Infer an aggregation function based on column name and type"""
    # Don't aggregate certain columns; ignore them
    last = get_last(prop)
    if last in ['x_root', 'x_contained_by_ref', 'type', 'id']:
        return None

    if prop == 'number_observed':
        return ('SUM', prop, prop)
    elif prop in ['first_observed', 'start']:
        return ('MIN', prop, prop)
    elif prop in ['last_observed', 'end']:
        return ('MAX', prop, prop)

    if ((sco_type == 'network-traffic' and prop.endswith('_port'))
        or (sco_type == 'process' and prop.endswith('pid'))):
        func = 'NUNIQUE'
        alias = f'unique_{prop}'
    elif col_type.lower() in ['integer', 'bigint']:
        func = 'AVG'
        alias = f'mean_{prop}'
    else:
        func = 'NUNIQUE'
        alias = f'unique_{prop}'

    return (func, prop, alias)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'ijson',
    'lark-parser',
    'orjson>=3.3.1',
    'psycopg2-binary',
    'tabulate',
    'typer',
    'ujson'
]

setup_requirements = ['pytest-runner', 'wheel']

test_requirements = ['pytest>=3', ]

setup(
    author="IBM Security",
    author_email='pcoccoli@us.ibm.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Columnar storage for STIX 2.0 observations.",
    entry_points={
        'console_scripts': [
            'firepit=firepit.cli:app',
            'splint=firepit.splint:app',
        ],
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='firepit',
    name='firepit',
    packages=find_packages(include=['firepit', 'firepit.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/opencybersecurityalliance/firepit',
    version='1.3.6',
    zip_safe=False,
)



# type: ignore[attr-defined]

"""
Typer-based CLI for STIX processing and linting
"""

import datetime
import logging
import json
import os
import sys
import uuid

import typer

from firepit import raft


logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


app = typer.Typer(
    name="splint",
    help="STIX processing and linting",
    add_completion=False,
)


TIME_FMT = '%Y-%m-%dT%H:%M:%S.%f'


def _timefmt(ts):
    return ts.strftime(TIME_FMT)[:-3] + 'Z'


def _start_bundle():
    bundle_id = 'bundle--' + str(uuid.uuid4())
    sys.stdout.write('{"type":"bundle",'
                     '"id": "')
    sys.stdout.write(bundle_id + '",')
    sys.stdout.write('"spec_version":"2.0",'
                     '"objects":[')


def _end_bundle():
    sys.stdout.write(']}')


@app.command()
def randomize_ids(
    filename: str = typer.Argument(..., help="STIX bundle file"),
):
    """Randomize STIX observation IDs in a bundle"""
    _start_bundle()
    count = 0
    ds_id = None
    for obj in raft.get_objects(filename):
        if 'type' in obj:
            obj_type = obj['type']
            new_id = f'{obj_type}--{uuid.uuid4()}'
            if obj_type == 'identity' and not ds_id:
                ds_id = new_id
            else:
                assert ds_id, 'No identity object in bundle?'
                obj['created_by_ref'] = ds_id
            obj['id'] = new_id
            if 'modified' in obj:
                obj['modified'] = _timefmt(datetime.datetime.utcnow())
        blob = json.dumps(obj, separators=(',', ':'))
        if count:
            sys.stdout.write(f',{blob}')
        else:
            sys.stdout.write(f'{blob}')
        count += 1
    _end_bundle()


@app.command()
def dedup_ids(
    filename: str = typer.Argument(..., help="STIX bundle file"),
):
    """Replace duplicate IDs with random IDs"""
    _start_bundle()
    count = 0
    ds_id = None
    ds_id_changed = False
    ids = set()
    for obj in raft.get_objects(filename):
        if 'type' in obj:
            old_id = obj.get('id', '')
            obj_type = obj['type']
            modified = False
            if old_id in ids:
                new_id = f'{obj_type}--{uuid.uuid4()}'
                obj['id'] = new_id
                modified = True
                ids.add(new_id)
            else:
                ids.add(old_id)
            if obj_type == 'identity' and not ds_id:
                ds_id = obj['id']
                if modified:
                    ds_id_changed = True
            elif ds_id_changed:
                assert ds_id, 'No identity object in bundle?'
                obj['created_by_ref'] = ds_id
                modified = True
            if 'modified' in obj and modified:
                obj['modified'] = _timefmt(datetime.datetime.utcnow())
        blob = json.dumps(obj, separators=(',', ':'))
        if count:
            sys.stdout.write(f',{blob}')
        else:
            sys.stdout.write(f'{blob}')
        count += 1
    _end_bundle()


@app.command()
def limit(
    n: int = typer.Argument(..., help="Max number of observations"),
    filename: str = typer.Argument(..., help="STIX bundle file"),
):
    """Truncate STIX bundle"""
    _start_bundle()
    count = 0
    for obj in raft.get_objects(filename):
        if count > n:
            break
        blob = json.dumps(obj, separators=(',', ':'))
        if count:
            sys.stdout.write(f',{blob}')
        else:
            sys.stdout.write(f'{blob}')
        count += 1
    _end_bundle()


if __name__ == "__main__":
    app()



import os
from firepit import get_storage

def tmp_storage(tmpdir, clear=True):
    dbname = os.getenv('FIREPITDB', str(tmpdir.join('test.db')))
    session = os.getenv('FIREPITID', 'test-session')

    if clear:
        # Clear out previous test session
        store = get_storage(dbname, session)
        store.delete()

    return get_storage(dbname, session)
